	/* vim: set expandtab shiftwidth=4 softtabstop=4 tabstop=4: */

	/**
	 * gRapid
	 *
	 * gRapid is a convenient RapidShare downloading tool (for premium users)
	 *
	 * (C) Copyright 2007 Mukul Sabharwal [http://mjsabby.com]
	 *	   All Rights Reserved
	 *
	 * @link http://code.google.com/p/grapid
	 * @author Mukul Sabharwal <mjsabby@gmail.com>
	 * @version $Id: gRapid.java,v 1.4 2007/05/27 00:27:30 mjsabby Exp $
	 * @copyright Copyright &copy; 2007 Mukul Sabharwal
	 * @license http://www.gnu.org/copyleft/gpl.html
	 * @package gRapid
	 */

	import java.io.*;
	import java.net.*;
	import java.util.*;
	import javax.net.ssl.*;
	import java.util.regex.*;

	/**
	 * Download manager class
	 * @author mjsabby
	 */
	class gRapidDownloadManager extends Thread {

		/**
		 * Lock object
		 */
		static Object lock;

		/**
		 * Premium cookie
		 */
		static String cookie;

		/**
		 * File list
		 */
		Stack<String> fileList;

		/**
		 * gRapidDownloadManager thread
		 */
		static Thread parentThread;

		/**
		 * Counter for wget log files
		 */
		static int counter;

		/**
		 * Maximum allowed simultaneous downloads
		 */
		static int maxSimultaneousDownloads;

		/**
		 * Current simultaneous downloads running
		 */
		static int curSimultaneousDownloads;

		/**
		 * Is the parent thread sleeping?
		 */
		static boolean parentThreadSleeping;

		/**
		 * Constructor
		 * @param f List of files to download
		 * @param c Cookie
		 * @param m Maximum parallel downloads to queue
		 */
		gRapidDownloadManager(Stack<String> f, String c, int m)
		{
			lock = new Object();
			cookie = c;
			fileList = f;
			maxSimultaneousDownloads = m;
			curSimultaneousDownloads = 0;
		}

		/**
		 * The download manager
		 */
		public void run()
		{
			parentThread = Thread.currentThread();
			while (true)
			{
				synchronized(lock)
				{
					while (fileList.size() > 0 && curSimultaneousDownloads < maxSimultaneousDownloads)
					{
							curSimultaneousDownloads++;
							(new gRapidDownload(fileList.pop(), cookie)).start();
					}
				}

				if (fileList.size() < 1) return; // no more downloads in the queue, terminate thread

				try {
					synchronized(lock) { parentThreadSleeping = true; }
					Thread.sleep(999999999); // sleep for a really long time!
				} catch (InterruptedException e) {
					synchronized(lock) { parentThreadSleeping = false; }
				}
			}
		}

		/**
		 * Individual download thread
		 * @author mjsabby
		 */
		private class gRapidDownload extends Thread {

			/**
			 * Url to download
			 */
			private String url;

			/**
			 * Premium cookie
			 */
			private String cookie;

			/**
			 * Constructor
			 * @param u Url
			 * @param c Cookie
			 */
			gRapidDownload(String u, String c)
			{
				url = u;
				cookie = c;
			}

			/**
			 * The actual wget download thread
			 */
			public void run()
			{
				try {
					Runtime rt = Runtime.getRuntime();
					Process proc = rt.exec("wget -c --directory-prefix=Downloads --no-cookies --header \"Cookie: " + cookie + "\" --append-output=wget-log.txt " + url);
					InputStream in = proc.getInputStream();
					proc.waitFor();
					in.close();
				} catch (Exception e) {
					System.out.print(e.getLocalizedMessage());
				} finally {
					synchronized(lock) {
						try { Thread.sleep(100); } catch (Exception e) { }
						curSimultaneousDownloads--;
						if (parentThreadSleeping) parentThread.interrupt();
					}
				}
			}
		}
	}

	/**
	 * gRapid class
	 * @author mjsabby
	 */
	public class gRapid {

		/**
		 * Fetch the actual url and write to outfile
		 * @param urltxt URL
		 * @param cookie Cookie
		 * @return
		 */
		private static String fetch(String urltxt, String cookie)
		{
			try {
				// open the url
				URL url = new URL(urltxt);
				HttpURLConnection conn = (HttpURLConnection)url.openConnection();
				//conn.setRequestProperty("Cookie", cookie);

				// read entire stream
				InputStream source = url.openStream();
				String data = new Scanner(source).useDelimiter("\\A").next();

				// Match the FORM tag for the POST url
				Pattern p = Pattern.compile("form action=\"(.*)\" method=\"post\"");
				Matcher m = p.matcher(data);

				// Not found, probably bad link, forbidden or something
				if (!m.find()) return "";
				urltxt = m.group(1); // save back to url

				// open url to specify premium downloads
				url = new URL(urltxt);
				conn = (HttpURLConnection)url.openConnection();

				// send our cookie, and post data along
				conn.setRequestProperty("Cookie", cookie);
				conn.setRequestMethod("POST");
				conn.setDoOutput(true);

				// Send POST across the wire
				OutputStreamWriter out = new OutputStreamWriter(conn.getOutputStream());
				out.write("dl.start=PREMIUM");
				out.flush();

				// read server reply
				BufferedReader in = new BufferedReader(new InputStreamReader(conn.getInputStream()));
				StringBuffer sb = new StringBuffer();
				while ((data = in.readLine()) != null) sb.append(data + System.getProperty("line.separator"));

				// Get the real url or throw error
				data = urltxt.substring(urltxt.lastIndexOf("/")+1);
				p = Pattern.compile("<tr><td><a href=\"(.*?)" + data);
				m = p.matcher(sb.toString());

				data = (m.find()) ? (m.group(1) + data + System.getProperty("line.separator")) : "";
				return data;
			} catch (Exception e) { return ""; }
		}

		/**
		 * Returns cookie contents once logged into RapidShare
		 * @param username
		 * @param password
		 * @return cookie contents
		 */
		private static String saveCookie(String username, String password)
		{

			TrustManager[] trustAllCerts = new TrustManager[] {
				new X509TrustManager() {
					public java.security.cert.X509Certificate[] getAcceptedIssuers() {return null;}
					public void checkClientTrusted(java.security.cert.X509Certificate[] certs, String authType) {}
					public void checkServerTrusted(java.security.cert.X509Certificate[] certs, String authType) {}
				}
			};

			try {

				// disable certificate checking
				SSLContext sc = SSLContext.getInstance("SSL");
				sc.init(null, trustAllCerts, new java.security.SecureRandom());
				HttpsURLConnection.setDefaultSSLSocketFactory(sc.getSocketFactory());

				// init
				URL url = new URL("https://ssl.rapidshare.com/cgi-bin/premiumzone.cgi");
				HttpsURLConnection conn = (HttpsURLConnection)url.openConnection();

				// set POST method
				conn.setRequestMethod("POST");
				conn.setDoOutput(true);

				// send the "param" across the wire
				OutputStreamWriter out = new OutputStreamWriter(conn.getOutputStream());
				out.write("login=" + username + "&password=" + password);
				out.flush();

				// retrieve server's header response
				String cookie = conn.getHeaderField("Set-Cookie");
				cookie = cookie.substring(0, cookie.indexOf(";"));

				// cleanup
				out.close();
				conn.disconnect();

				// return
				return cookie;
			} catch (Exception e) { return ""; }
		}

		/**
		 * Main method
		 * @param argv
		 * @throws Exception
		 */
		public static void main(String argv[]) throws Exception
		{
			// check for argument length
			if(argv.length != 4) {
				System.out.println("Usage: java -jar gRapid.jar username password input_file simultaneous_downloads");
				System.exit(0);
			}

			// Save the cookie
			String cookie = saveCookie(argv[0], argv[1]);
			if (cookie.length() < 6) { // invalid username / password
				System.out.println("Authentication failure");
				System.exit(0);
			}

			// I/O initialize
			Scanner in = new Scanner(new File(argv[2]));
			BufferedWriter out = new BufferedWriter(new FileWriter("out.txt"));

			// create an empty Stack for our files
			Stack<String> files = new Stack<String>();
			Stack<String> nList = new Stack<String>();

			// Read url's; write url's
			while (in.hasNextLine())
			{
				files.push(in.nextLine().trim());
				nList.push(fetch(files.peek(), cookie));
				out.write(nList.peek());
				out.flush();
			}

			// I/O cleanup
			in.close();
			out.close();

			// start download manager
			new gRapidDownloadManager(nList, cookie, Integer.parseInt(argv[3])).start();
		}
	}